#!/usr/bin/env python3

import argparse
import os
import re
from glob import glob
from tqdm import tqdm

from ucca.convert import to_text
from ucca.ioutil import file2passage, get_passages_with_progress_bar

desc = """Parses files in UCCA standard format, and writes as text files or a text file with a line per passage."""


def numeric(x):
    try:
        return tuple(map(int, re.findall("\d+", x)))
    except ValueError:
        return x


def write_text(passage, f, lang):
    for line in to_text(passage, lang=lang):
        print(line, file=f)


def main(args):
    os.makedirs(args.outdir, exist_ok=True)
    if args.join:
        with open(os.path.join(args.outdir, args.join), "w", encoding="utf-8") as f:
            for passage in get_passages_with_progress_bar(sorted(args.filenames, key=numeric), desc="Converting"):
                write_text(passage, f, lang=args.lang)
    else:  # one file per passage
        for pattern in args.filenames:
            for filename in tqdm(glob(pattern) or [pattern], desc="Converting", unit=" passages"):
                passage = file2passage(filename)
                basename = os.path.splitext(os.path.basename(filename))[0]
                with open(os.path.join(args.outdir, basename + ".txt"), "w", encoding="utf-8") as f:
                    write_text(passage, f, lang=args.lang)
    print("Wrote '%s'." % args.outfile)


if __name__ == "__main__":
    argparser = argparse.ArgumentParser(description=desc)
    argparser.add_argument("filenames", nargs="+", help="passage file names to convert")
    argparser.add_argument("-o", "--outdir", default=".", help="output directory")
    argparser.add_argument("-l", "--lang", default="en", help="language two-letter code for sentence model")
    argparser.add_argument("-j", "--join", help="write just one text file with this name, with one line per passage")
    main(argparser.parse_args())



#!/usr/bin/env python3

import sys
from itertools import count

import argparse
import os

from ucca.convert import split2sentences, split_passage
from ucca.ioutil import passage2file, get_passages_with_progress_bar, external_write_mode
from ucca.normalization import normalize
from ucca.textutil import extract_terminals

desc = """Parses XML files in UCCA standard format, and writes a passage per sentence."""


class Splitter:
    def __init__(self, sentences, enum=False, suffix_format="%03d", suffix_start=0):
        self.sentences = sentences
        self.sentence_to_index = dict(map(reversed, enumerate(sentences)))
        self.enumerate = enum
        self.suffix_format = suffix_format
        self.suffix_start = suffix_start
        self.index = 0

    @classmethod
    def read_file(cls, filename, **kwargs):
        if filename is None:
            return None
        with open(filename, encoding="utf-8") as f:
            sentences = [l.strip() for l in f]
        return cls(sentences, **kwargs)

    def split(self, passage):
        ends = []
        ids = []
        tokens = []
        for terminal in extract_terminals(passage):
            tokens.append(terminal.text)
            sentence = " ".join(tokens)
            # if len(tokens) > max(map(len, map(str.split, sentence_to_index))):
            #     raise ValueError("Failed matching '%s'" % sentence)
            if self.index is not None and self.index < len(self.sentences) and \
                    self.sentences[self.index].startswith(sentence):  # Try matching next sentence rather than shortest
                index = self.index if self.sentences[self.index] == sentence else None
            else:
                index = self.index = self.sentence_to_index.get(sentence)
            if index is not None:
                ends.append(terminal.position)
                ids.append(str(index))
                tokens = []
                self.index += 1
        return split_passage(passage, ends, ids=ids if self.enumerate else None,
                             suffix_format=self.suffix_format, suffix_start=self.suffix_start)


def main(args):
    splitter = Splitter.read_file(args.sentences, enum=args.enumerate,
                                  suffix_format=args.suffix_format, suffix_start=args.suffix_start)
    os.makedirs(args.outdir, exist_ok=True)
    i = 0
    for passage in get_passages_with_progress_bar(args.filenames, "Splitting"):
        for sentence in splitter.split(passage) if splitter else split2sentences(
                passage, remarks=args.remarks, lang=args.lang, ids=map(str, count(i)) if args.enumerate else None):
            i += 1
            outfile = os.path.join(args.outdir, args.prefix + sentence.ID + (".pickle" if args.binary else ".xml"))
            with external_write_mode():
                print("Writing passage file for sentence '%s'..." % outfile, file=sys.stderr)
            if args.normalize:
                normalize(sentence)
            passage2file(sentence, outfile, binary=args.binary)


if __name__ == "__main__":
    argparser = argparse.ArgumentParser(description=desc)
    argparser.add_argument("filenames", nargs="+", help="passage file names to convert")
    argparser.add_argument("-o", "--outdir", default=".", help="output directory")
    argparser.add_argument("-p", "--prefix", default="", help="output filename prefix")
    argparser.add_argument("-f", "--suffix-format", default="%03d", help="sentence number suffix format")
    argparser.add_argument("-i", "--suffix-start", type=int, default=0, help="start index for number suffix")
    argparser.add_argument("-r", "--remarks", action="store_true", help="annotate original IDs")
    argparser.add_argument("-l", "--lang", default="en", help="language two-letter code for sentence model")
    argparser.add_argument("-b", "--binary", action="store_true", help="write in pickle binary format (.pickle)")
    argparser.add_argument("-s", "--sentences", help="optional input file with sentence at each line to split by")
    argparser.add_argument("-e", "--enumerate", action="store_true", help="set each output sentence ID by global order")
    argparser.add_argument("-N", "--no-normalize", dest="normalize", action="store_false",
                           help="do not normalize passages after splitting")
    main(argparser.parse_args())



#!/usr/bin/env python3
import sys

import argparse
from tqdm import tqdm

from uccaapp.api import ServerAccessor

desc = """Convert a passage file to JSON format and upload to UCCA-App as a completed task"""


class AnnotationTaskCreator(ServerAccessor):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def create_tasks(self, filename, log=None, **kwargs):
        log_h = open(log, "w", encoding="utf-8") if log else None
        lines = list(self.read_lines(filename))
        for user_id, task_id in tqdm(lines, unit="task", desc="Creating tasks"):
            task = self.create_task(**self.build_task(user_id, task_id, **kwargs))
            if log:
                print(task["id"], file=log_h, sep="\t", flush=True)
        print("Uploaded %d tasks successfully." % len(lines), file=sys.stderr)
        if log:
            log_h.close()

    def build_task(self, user_id, task_id, review=False, manager_comment=None, **kwargs):
        del kwargs
        user = self.get_task(user_id)
        task = self.get_task(task_id)
        assert task["type"] in (["ANNOTATION", "REVIEW"] if review else ["TOKENIZATION"]), \
            "Wrong input task given: %s for task ID %s" % (task["type"], task_id)
        assert task["status"] == "SUBMITTED", "Parent task is not submitted: %s" % task_id
        return dict(type="REVIEW" if review else "ANNOTATION", project=task["project"], user=user,
                    passage=task["passage"], manager_comment=manager_comment or task.get("manager_comment", ""),
                    user_comment=task.get("user_comment", ""), parent=task, is_demo=False, is_active=True)

    @staticmethod
    def read_lines(filename):
        with open(filename, encoding="utf-8") as f:
            for line in f:
                fields = line.strip().split()
                try:
                    user_id, task_id = fields
                except ValueError:
                    print("Error in line: " + line.strip(), file=sys.stderr)
                    continue
                yield user_id, task_id

    @staticmethod
    def add_arguments(argparser):
        argparser.add_argument("filename", help="a file where each line is a <User ID> <INPUT TASK ID>, "
                                                "where the input task may be an annotation/review task "
                                                "(if given --review) or a tokenization task")
        ServerAccessor.add_arguments(argparser)
        argparser.add_argument("-r", "--review", action="store_true", help="Create annotation/review task")
        argparser.add_argument("-l", "--log", help="filename to write log of uploaded passages to")
        argparser.add_argument("--manager-comment", help="Manager comment to set for all tasks")


def main(**kwargs):
    AnnotationTaskCreator(**kwargs).create_tasks(**kwargs)


if __name__ == "__main__":
    argument_parser = argparse.ArgumentParser(description=desc)
    AnnotationTaskCreator.add_arguments(argument_parser)
    main(**vars(argument_parser.parse_args()))
    sys.exit(0)



from itertools import groupby

import string
from operator import attrgetter

from ucca import layer0, layer1
from ucca.layer0 import NodeTags as L0Tags
from ucca.layer1 import EdgeTags as ETags, NodeTags as L1Tags

LINKAGE = (ETags.LinkArgument, ETags.LinkRelation)


def validate(passage, linkage=True, multigraph=False):
    for node in passage.layer(layer0.LAYER_ID).all:
        yield from NodeValidator(node).validate_terminal()
    heads = list(passage.layer(layer1.LAYER_ID).heads)
    found_linkage = False
    for node in heads:
        if node.tag == L1Tags.Linkage:
            found_linkage = True
        yield from NodeValidator(node).validate_top_level()
    stack = [heads]
    visited = set()
    path = []
    path_set = set(path)
    while stack:
        for node in stack[-1]:
            if node in path_set:
                yield "Detected cycle (%s)" % "->".join(n.ID for n in path)
            elif node not in visited:
                visited.add(node)
                path.append(node)
                path_set.add(node)
                stack.append(node.children)
                yield from NodeValidator(node).validate_non_terminal(linkage=linkage and found_linkage,
                                                                     multigraph=multigraph)
                break
        else:
            if path:
                path_set.remove(path.pop())
            stack.pop()


class NodeValidator:
    def __init__(self, node):
        self.node = node
        self.node_id = self.node.ID
        tree_id = self.node.extra.get("tree_id")
        if tree_id:
            self.node_id += ", %s" % tree_id
        self.incoming = tag_to_edge(node.incoming)
        self.outgoing = tag_to_edge(node)
        self.incoming_tags = set(self.incoming)
        self.outgoing_tags = set(self.outgoing)

    def validate_terminal(self):
        if not self.node.text:
            yield "Empty terminal text (%s)" % self.node_id
        if set(self.node.text).intersection(string.whitespace):
            yield "Whitespace in terminal text (%s): '%s'" % (self.node_id, self.node)
        if not self.incoming:
            yield "Orphan %s terminal (%s) '%s'" % (self.node.tag, self.node_id, self.node)
        elif len(self.node.incoming) > 1:
            yield "Reentrant %s terminal (%s) '%s'" % (self.node.tag, join(self.node.incoming), self.node)

    def validate_top_level(self):
        if self.node not in self.node.layer.heads and self.node.tag != L1Tags.Linkage:
            yield "Extra root (%s)" % self.node_id
        terminals = [n for n in self.node.children if n.layer.ID == layer0.LAYER_ID]
        if terminals:
            yield "Terminal children (%s) of root (%s)" % (join(terminals), self.node_id)
        s = self.outgoing_tags.difference((ETags.ParallelScene, ETags.Linker, ETags.Function, ETags.Ground,
                                           ETags.Punctuation, ETags.LinkRelation, ETags.LinkArgument))
        if s:
            yield "Top-level unit (%s) with %s children: %s" %\
                  (self.node_id, join(s), join(e.child for e in self.node if e.tag in s))

    def validate_non_terminal(self, linkage=False, multigraph=False):
        if linkage and self.node.tag == L1Tags.Linkage:
            yield from self.validate_linkage()
        elif self.node.tag == L1Tags.Foundational:
            yield from self.validate_foundational()
        primary_incoming = [e for e in self.node.incoming if not e.attrib.get("remote") and e.tag not in LINKAGE]
        if len(primary_incoming) > 1:
            yield "Unit (%s) with multiple non-remote parents (%s)" % (self.node_id, join(primary_incoming))
        remote_incoming = [e for e in self.node.incoming if e.attrib.get("remote")]
        if remote_incoming and not primary_incoming:
            yield "Unit (%s) with remote parents but no primary parents" % self.node_id
        for edge in self.node:
            if (edge.tag == ETags.Punctuation) != (edge.child.tag == L1Tags.Punctuation):
                yield "%s edge (%s) with %s child" % (edge.tag, edge, edge.child.tag)
            # FN parent of Punctuation is disallowed unless the FN is unanalyzable
            if (self.node.tag == L1Tags.Foundational) and (edge.child.tag == L0Tags.Punct) and \
                    not len(self.node.terminals) + len(self.node.punctuation) == len(self.node.children) > 1 or \
                    (self.node.tag == L1Tags.Punctuation) and not (edge.child.tag == L0Tags.Punct):
                yield "%s unit (%s) with %s child (%s)" % (self.node.tag, self.node_id, edge.child.tag, edge.child.ID)
        if self.node.attrib.get("implicit"):
            if self.node.outgoing:
                yield "Implicit unit (%s) with children (%s)" % (self.node_id, join(self.node.children))
        elif self.node.tag in (L1Tags.Foundational, L1Tags.Linkage, L1Tags.Punctuation) and \
                all(e.attrib.get("remote") for e in self.node):
            yield "Non-implicit unit (%s) with no primary children" % (self.node_id)
        for tag in (ETags.Function, ETags.ParallelScene, ETags.Linker, ETags.LinkRelation,
                    ETags.Connector, ETags.Punctuation, ETags.Terminal):
            s = self.incoming.get(tag, ())
            if len(s) > 1:
                yield "Unit (%s) with multiple %s parents (%s)" % (self.node_id, tag, join(e.parent for e in s))
        for tag in (ETags.LinkRelation, ETags.Process, ETags.State):
            s = self.outgoing.get(tag, ())
            if len(s) > 1:
                yield "Unit (%s) with multiple %s children (%s)" % (self.node_id, tag, join(e.child for e in s))
        if ETags.Function in self.incoming:
            s = self.outgoing_tags.difference((ETags.Terminal, ETags.Punctuation))
            if s:
                yield "%s unit (%s) with %s children: %s" % (ETags.Function, self.node_id, join(s), self.node)
        if ETags.Linker in self.incoming_tags and linkage and ETags.LinkRelation not in self.incoming_tags:
            yield "%s unit (%s) with no incoming %s" % (ETags.Linker, self.node_id, ETags.LinkRelation)
        if not multigraph:
            key = attrgetter("child.ID")
            for child_id, edges in groupby(sorted(self.node, key=key), key=key):
                edges = list(edges)
                if len(edges) > 1:
                    yield "Multiple edges from %s to %s: %s" % (self.node_id, child_id, ", ".join(
                        "%d %s" % (len(e), t) for t, e in tag_to_edge(edges).items()))

    def validate_linkage(self):
        if self.node.incoming:
            yield "Non-root %s unit (%s)" % (self.node.tag, self.node_id)
        s = self.outgoing_tags.difference(LINKAGE)
        if s:
            yield "%s unit (%s) with %s children" % (self.node.tag, self.node_id, join(s))
        if ETags.LinkRelation not in self.outgoing:
            yield "%s unit without %s child" % (self.node.tag, ETags.LinkRelation)

    def validate_foundational(self):
        if self.node.participants and not self.node.is_scene():
            yield "Unit (%s) with participants but without main relation: %s" % (self.node_id, self.node)
        if self.node.process and self.node.state:
            yield "Unit (%s) with both process (%s) and state (%s)" % (self.node_id, self.node.process, self.node.state)
        if self.node.parallel_scenes:
            s = self.outgoing_tags.difference((ETags.ParallelScene, ETags.Punctuation, ETags.Linker,
                                               ETags.Ground, ETags.Relator, ETags.Function))
            if s:
                yield "Unit (%s) with parallel scenes has %s children: %s" %\
                      (self.node_id, join(s), join(e.child for e in self.node if e.tag in s))
        s = self.outgoing_tags.intersection(LINKAGE)
        if s:
            yield "Non-linkage unit (%s) with %s children: %s" %\
                  (self.node_id, join(s), join(e.child for e in self.node if e.tag in s))


def tag_to_edge(edges):
    d = {}
    for edge in edges:
        d.setdefault(edge.tag, []).append(edge)
    return d


def join(items):
    return ", ".join(map(str, items))



#!/usr/bin/env python3
import sys

import argparse

from uccaapp.api import ServerAccessor

desc = """Download categories from one UCCA-App server and upload to another UCCA-App server"""


def add_arguments(argparser):
    argparser.add_argument("category-ids", nargs="+", type=int, help="IDs of tasks to export and import")
    argparser.add_argument("--server-address-orig", required=True, help="UCCA-App origin server")
    argparser.add_argument("--email-orig", help="UCCA-App origin email")
    argparser.add_argument("--password-orig", help="UCCA-App origin password")
    argparser.add_argument("--server-address-target", required=True, help="UCCA-App target server")
    argparser.add_argument("--email-target", help="UCCA-App target email")
    argparser.add_argument("--password-target", help="UCCA-App target password")
    argparser.add_argument("-v", "--verbose", action="store_true", help="detailed output")


def main(args):
    server_accessor_origin = ServerAccessor(server_address=args.server_address_orig,
                                            email=args.email_orig, password=args.password_orig,
                                            verbose=args.verbose)
    server_accessor_target = ServerAccessor(server_address=args.server_address_target,
                                            email=args.email_target, password=args.password_target,
                                            verbose=args.verbose)
    for category_id in args.category_ids:
        category_out = server_accessor_origin.get_category(category_id)
        server_accessor_target.create_category(**category_out)


if __name__ == "__main__":
    argument_parser = argparse.ArgumentParser(description=desc)
    add_arguments(argument_parser)
    main(argument_parser.parse_args())
    sys.exit(0)



import sys
from multiprocessing import Pool

import argparse

from ucca.ioutil import get_passages_with_progress_bar, external_write_mode
from ucca.normalization import normalize
from ucca.validation import validate


class Validator:
    def __init__(self, normalization=False, extra=False, linkage=True, multigraph=False, strict=False):
        self.normalization = normalization
        self.extra = extra
        self.linkage = linkage
        self.multigraph = multigraph
        self.strict = strict

    def validate_passage(self, passage):
        if self.normalization:
            normalize(passage, extra=self.extra)
        errors = list(validate(passage, linkage=self.linkage, multigraph=self.multigraph))
        passage_id = passage.ID
        user_id = passage.attrib.get("userID")
        if user_id:
            passage_id += " " + user_id
        if self.strict:
            print_errors(passage_id, errors)
        return passage_id, errors


def main(args):
    validator = Validator(args.normalize, args.extra, linkage=args.linkage, multigraph=args.multigraph,
                          strict=args.strict)
    with Pool(10) as pool:
        errors = pool.map(validator.validate_passage,
                          get_passages_with_progress_bar(args.filenames, desc="Validating", converters={}))
    errors = dict((k, v) for k, v in errors if v)
    if errors:
        if not args.strict:
            id_len = max(map(len, errors))
            for passage_id, es in sorted(errors.items()):
                print_errors(passage_id, es, id_len)
        sys.exit(1)
    else:
        print("No errors found.")


def print_errors(passage_id, errors, id_len=None):
    for i, e in enumerate(errors):
        with external_write_mode():
            print("%-*s|%s" % (id_len or len(passage_id), "" if i else passage_id, e), flush=True)


def check_args(parser, args):
    if args.extra and not args.normalize:
        parser.error("Cannot specify --extra without --normalize")
    return args


if __name__ == "__main__":
    argparser = argparse.ArgumentParser(description="Validate UCCA passages")
    argparser.add_argument("filenames", nargs="+", help="files or directories to validate")
    argparser.add_argument("-S", "--strict", action="store_true", help="fail as soon as a violation is found")
    argparser.add_argument("-n", "--normalize", action="store_true", help="normalize before validation")
    argparser.add_argument("-e", "--extra", action="store_true", help="extra normalization rules")
    argparser.add_argument("--no-linkage", dest="linkage", action="store_false", help="skip linkage validations")
    argparser.add_argument("--multigraph", action="store_true", help="allow multiple edges with the same parent+child")
    main(check_args(argparser, argparser.parse_args()))



import csv
import os
from argparse import ArgumentParser

from ucca import layer1
from ucca.ioutil import get_passages_with_progress_bar, write_passage
from ucca.normalization import destroy, copy_edge

desc = """Convert the English Wiki corpus from version 2.0 to 1.2"""


def replace_time_and_quantifier(edge):
    if edge.tag in (layer1.EdgeTags.Time, layer1.EdgeTags.Quantifier):
        edge.tag = layer1.EdgeTags.Adverbial if edge.parent.is_scene() else layer1.EdgeTags.Elaborator
        if len(edge.parent.parents) == 1 and edge.parent.incoming[0].tag == edge.tag:
            for e in edge.parent:
                copy_edge(e, parent=edge.parent.parents[0])
            destroy(edge.parent)
        return True
    return False


RULES = (replace_time_and_quantifier,)


def convert_passage(passage, report_writer):
    for rule in RULES:
        for node in passage.layer(layer1.LAYER_ID).all:
            for edge in node:
                parent = edge.parent
                parent_str = str(parent)
                if rule(edge):
                    report_writer.writerow((rule.__name__, passage.ID, edge, parent_str, parent))


def main(args):
    os.makedirs(args.outdir, exist_ok=True)
    with open(args.outfile, "w", encoding="utf-8", newline="") as f:
        writer = csv.writer(f)
        writer.writerow(("rule", "passage", "edge", "before", "after"))
        for passage in get_passages_with_progress_bar(args.passages, desc="Converting"):
            convert_passage(passage, report_writer=writer)
            write_passage(passage, outdir=args.outdir, prefix=args.prefix, verbose=args.verbose)
            f.flush()
    print("Wrote '%s'" % args.outfile)


if __name__ == "__main__":
    argparser = ArgumentParser(description=desc)
    argparser.add_argument("passages", nargs="+", help="the corpus, given as xml/pickle file names")
    argparser.add_argument("-o", "--outdir", default=".", help="output directory")
    argparser.add_argument("-p", "--prefix", default="", help="output filename prefix")
    argparser.add_argument("-O", "--outfile", default=os.path.splitext(argparser.prog)[0] + ".csv", help="log file")
    argparser.add_argument("-v", "--verbose", action="store_true", help="print more information")
    main(argparser.parse_args())



import argparse
import sys

from ucca import layer1
from uccaapp.download_task import TaskDownloader

desc = "Get all units according to a specified filter. Units that meet any of the filters are output."

def get_top_level_ancestor(node):
    """
    Traverses the passage upwards until a unit which is immediately below the root is reached
    :param node:
    :return:
    """
    # if node is already the root, return it
    if not node.fparent:
        return node
    else:
        parent = node
        while parent.fparent.fparent:
            parent = parent.fparent
        return parent

def main(output = None, comment = False, categories = (), **kwargs):
    filtered_nodes = []
    for passage, task_id, user_id in TaskDownloader(**kwargs).download_tasks(**kwargs, write=False):
        for node in passage.layer(layer1.LAYER_ID).all:
            if comment and node.extra.get("remarks"):
                filtered_nodes.append(("comment",node,task_id,user_id))
            else:
                all_tags = node.extra.get("all_tags")
                if all_tags:
                    intersection = set(categories) & set(all_tags.split(';'))
                    if intersection:
                        filtered_nodes.append((str(list(intersection)),node,task_id,user_id))

    if output:
        with open(output,'w') as f:
            for filter_type,node,task_id,user_id in filtered_nodes:
                ancestor = get_top_level_ancestor(node)
                print(filter_type, task_id, user_id, node.extra.get("tree_id"), node.to_text(),
                      ancestor, str(node.extra.get("remarks")).replace("\n","|"), file=f, sep="\t")

if __name__ == "__main__":
    argument_parser = argparse.ArgumentParser(description=desc)
    TaskDownloader.add_arguments(argument_parser)
    argument_parser.add_argument("--output", help="output file name")
    argument_parser.add_argument("--categories", nargs="+", default=(), help="Abbreviations of the names of the categories to filter by")
    argument_parser.add_argument("--comment", action="store_true", help="Output all the units that have comments")

    main(**vars(argument_parser.parse_args()))



#!/usr/bin/env python3
import sys

import argparse
from tqdm import tqdm

from uccaapp.api import ServerAccessor

desc = """Set the external ID for passages"""


class ExternalIdSetter(ServerAccessor):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def set_external_ids(self, filename, **kwargs):
        del kwargs
        with open(filename, encoding="utf-8") as f:
            passage_id_to_external_id = list(map(str.split, map(str.strip, f)))
        for external_id, passage_id in tqdm(passage_id_to_external_id, unit=" passages", desc="Setting external IDs"):
            passage = self.get_passage(passage_id)
            passage["external_id"] = external_id
            passage_out = self.update_passage(**passage)
            assert passage_out["external_id"] == external_id, "External ID failed to update for passage %s" % passage_id
            yield passage_out

    @staticmethod
    def add_arguments(argparser):
        argparser.add_argument("filename", help="file with lines of the form <EXTERNAL ID> <PASSAGE ID>")
        ServerAccessor.add_arguments(argparser)


def main(**kwargs):
    list(ExternalIdSetter(**kwargs).set_external_ids(**kwargs))


if __name__ == "__main__":
    argument_parser = argparse.ArgumentParser(description=desc)
    ExternalIdSetter.add_arguments(argument_parser)
    main(**vars(argument_parser.parse_args()))
    sys.exit(0)



#!/usr/bin/env python3
import sys

import argparse

from ucca.convert import from_text, to_json
from uccaapp.api import ServerAccessor

desc = """Upload a passage from a streussel format file"""


class StreusselPassageUploader(ServerAccessor):
    def __init__(self, user_id, source_id, project_id, **kwargs):
        super().__init__(**kwargs)
        self.set_source(source_id)
        self.set_project(project_id)
        self.set_user(user_id)

    def upload_streussel_passage_file(self, filenames, log=None, **kwargs):
        del kwargs
        log_h = open(log, "w", encoding="utf-8") if log else None
        with open(filenames) as f_all:
            for filename in f_all:
                passage_text = ""
                external_id = "None given"
                filename = filename.strip()
                with open(filename, encoding="utf-8") as f:
                    for line in f:
                        line = line.strip()
                        if not line:
                            continue
                        elif line.startswith("#"):
                            fields = line.split()
                            if len(fields) != 4 or fields[1] != "sent_id":
                                print("FORMAT ERROR in " + filename, file=sys.stderr)
                            else:
                                external_id = fields[3].split("-")[1]
                        else:
                            passage_text = passage_text + " " + line
                passage_out = self.create_passage(text=passage_text.strip(), external_id=external_id, type="PUBLIC",
                                                  source=self.source)
                task_in = dict(type="TOKENIZATION", status="SUBMITTED", project=self.project,
                               user=self.user, passage=passage_out, manager_comment="External ID: " + external_id,
                               user_comment="", parent=None, is_demo=False, is_active=True)
                tok_task_out = self.create_task(**task_in)
                tok_user_task_in = dict(tok_task_out)

                passage = list(from_text(passage_text.split(), tokenized=True))[0]
                tok_user_task_in.update(to_json(passage, return_dict=True, tok_task=True))

                self.submit_tokenization_task(**tok_user_task_in)
                print("Uploaded passage " + filename + " successfully.", file=sys.stderr)
                if log:
                    print(filename.split(".")[-2], passage_out["id"], tok_task_out["id"], file=log_h, sep="\t")
        if log:
            log_h.close()

    @staticmethod
    def add_arguments(argparser):
        argparser.add_argument("filenames", help="passage file names to convert and upload")
        argparser.add_argument("-l", "--log", help="filename to write log of uploaded passages to")
        ServerAccessor.add_project_id_argument(argparser)
        ServerAccessor.add_source_id_argument(argparser)
        ServerAccessor.add_user_id_argument(argparser)
        ServerAccessor.add_arguments(argparser)


def main(**kwargs):
    StreusselPassageUploader(**kwargs).upload_streussel_passage_file(**kwargs)


if __name__ == "__main__":
    argument_parser = argparse.ArgumentParser(description=desc)
    StreusselPassageUploader.add_arguments(argument_parser)
    main(**vars(argument_parser.parse_args()))
    sys.exit(0)



#!/usr/bin/env python3
import sys

import argparse
from tqdm import tqdm

from uccaapp.api import ServerAccessor

desc = """Get passage ID for tasks"""


class PassageIdGetter(ServerAccessor):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def get_passage_ids(self, filename, **kwargs):
        del kwargs
        with open(filename, encoding="utf-8") as f:
            task_ids = list(map(str.strip, f))
        for task_id in tqdm(task_ids, unit=" tasks", desc="Getting passage IDs"):
            task = self.get_task(task_id)
            passage_id = task["passage"]["id"]
            yield passage_id

    @staticmethod
    def add_arguments(argparser):
        argparser.add_argument("filename", help="file with lines of the form <TASK ID>")
        ServerAccessor.add_arguments(argparser)


def main(**kwargs):
    print(*PassageIdGetter(**kwargs).get_passage_ids(**kwargs), sep="\n")


if __name__ == "__main__":
    argument_parser = argparse.ArgumentParser(description=desc)
    PassageIdGetter.add_arguments(argument_parser)
    main(**vars(argument_parser.parse_args()))
    sys.exit(0)



import pytest

from ucca import layer1
from ucca.validation import validate
from .conftest import loaded, loaded_valid, multi_sent, crossing, discontiguous, l1_passage, empty, \
    create_passage, attach_terminals

"""Tests the validation module functions and classes."""


def unary_punct():
    p, l1, terms = create_passage(3, 3)
    h1 = l1.add_fnode(None, layer1.EdgeTags.ParallelScene)
    p1 = l1.add_fnode(h1, layer1.EdgeTags.Process)
    a1 = l1.add_fnode(h1, layer1.EdgeTags.Participant)
    l1.add_punct(h1, terms[2])
    attach_terminals(terms, p1, a1)
    return p


def binary_punct():
    p, l1, terms = create_passage(4, 3)
    h1 = l1.add_fnode(None, layer1.EdgeTags.ParallelScene)
    p1 = l1.add_fnode(h1, layer1.EdgeTags.Process)
    a1 = l1.add_fnode(h1, layer1.EdgeTags.Participant)
    l1.add_punct(h1, terms[2]).add(layer1.EdgeTags.Terminal, terms[3])
    attach_terminals(terms, p1, a1)
    return p


def unary_punct_under_fn():
    p, l1, terms = create_passage(3, 3)
    h1 = l1.add_fnode(None, layer1.EdgeTags.ParallelScene)
    p1 = l1.add_fnode(h1, layer1.EdgeTags.Process)
    a1 = l1.add_fnode(h1, layer1.EdgeTags.Participant)
    attach_terminals(terms, p1, a1, h1)
    return p


def punct_under_unanalyzable_fn():
    p, l1, terms = create_passage(3, 2)
    h1 = l1.add_fnode(None, layer1.EdgeTags.ParallelScene)
    p1 = l1.add_fnode(h1, layer1.EdgeTags.Process)
    attach_terminals(terms, p1, p1, p1)
    return p


@pytest.mark.parametrize("create, valid", (
        (loaded, False),
        (loaded_valid, True),
        (multi_sent, True),
        (crossing, True),
        (discontiguous, True),
        (l1_passage, True),
        (empty, False),
        (unary_punct, True),
        (binary_punct, True),
        (unary_punct_under_fn, False),
        (punct_under_unanalyzable_fn, True),
))
def test_evaluate_self(create, valid):
    p = create()
    errors = list(validate(p))
    if valid:
        assert not errors, p
    else:
        assert errors, p



import sys

import datetime
import psycopg2
from tqdm import tqdm
from xml.etree.ElementTree import tostring, fromstring as fromstring_xml

from ucca import convert
from ucca.ioutil import external_write_mode

UNK_LINKAGE_TYPE = 'UNK'
CONNECTION = None


def fromstring(text):
    text = text.replace(r"\u2019", "&apos;")
    text = text.replace(r"\u2013", "-")
    text = text.replace(r"\u2014", "-")
    text = text.replace(r"\u2032", "'")
    text = text.replace(r"\u201C", '"')
    text = text.replace(r"\u201D", '"')
    if r"\u" in text:
        raise Exception("Unescaped unicode: " + text)
    return fromstring_xml(text)


#######################################################################################
# Returns the most recent xmls from db with a passage id pid and usernames
# (a list). The xmls are ordered in the same way as the list usernames.
#######################################################################################
def get_xmls_by_username(host_name, db_name, username):
    c = get_cursor(host_name, db_name)
    uid = get_uid(host_name, db_name, username)
    c.execute("SELECT xml FROM xmls WHERE uid=%s AND ts IN (SELECT MAX(ts) from xmls GROUP BY paid)", (uid,))
    for queryset in c.fetchall():
        yield fromstring(queryset[0])


def get_xml_trees(host_name, db_name, pid, usernames=None, graceful=False):
    """ Params: db, host, paragraph id, the list of usernames wanted,
        Optional:
        graceful: True if no excpetions are to be raised
        excpetion raised if a user did not submit an annotation for the passage
        returns a list of xml roots elements
        """
    c = get_cursor(host_name, db_name)
    xmls = []
    if usernames is None:
        c.execute("SELECT xml FROM xmls WHERE paid=%s ORDER BY ts DESC", (pid,))
        queryset = c.fetchone()
        if queryset is not None:
            xmls.append(fromstring(queryset[0]))
    else:
        for username in usernames:
            username = str(username)  # precaution for cases bad input e.g. 101
            cur_uid = get_uid(host_name, db_name, username)
            c.execute("SELECT xml FROM xmls WHERE paid=%s AND uid=%s ORDER BY ts DESC", (pid, cur_uid))
            raw_xml = c.fetchone()
            if not raw_xml and not graceful:
                raise Exception("The user " + username +
                                " did not submit an annotation for this passage")
            else:
                xmls.append(fromstring(raw_xml[0]))
    return xmls


def get_by_xids(host_name, db_name, xids, **kwargs):
    """Returns the passages that correspond to xids (which is a list of them)"""
    del kwargs
    c = get_cursor(host_name, db_name)
    xmls = []
    for xid in xids:
        c.execute("SELECT xml FROM xmls WHERE id=%s", (int(xid),))
        queryset = c.fetchone()
        if queryset is None:
            raise Exception("The xid " + xid + " does not exist")
        else:
            xmls.append(fromstring(queryset[0]))
    return xmls


def get_most_recent_passage_by_uid(uid, passage_id, host_name, db_name, verbose=False, write_xids=None, strict=False,
                                   **kwargs):
    del kwargs
    c = get_cursor(host_name, db_name)
    uid = (uid,) if isinstance(uid, (str, int)) else tuple(uid)
    c.execute("SELECT xml,status,ts,id,uid FROM xmls WHERE uid IN %s AND paid = %s ORDER BY ts DESC", (uid, passage_id))
    queryset = c.fetchone()
    raw_xml, status, ts, xid, uid = 5 * [None]
    if queryset is None:
        if strict:
            raise Exception("The user %s did not annotate passage %s" % (uid, passage_id))
    else:
        raw_xml, status, ts, xid, uid = queryset
    if write_xids:
        with open(write_xids, "a") as f:
            print(passage_id, xid, uid, ts, file=f, sep="\t")
    if queryset is None:
        return None
    if int(status) != 1:  # if not submitted
        with external_write_mode():
            print("The most recent xml for uid %s and paid %s is not submitted." % (uid, passage_id), file=sys.stderr)
    if verbose:
        with external_write_mode():
            print("Timestamp: %s, uid: %d, xid: %d" % (ts, uid, xid))
    return fromstring(raw_xml)


def get_uid(host_name, db_name, username):
    """Returns the uid matching the given username."""
    c = get_cursor(host_name, db_name)
    c.execute("SELECT id FROM users WHERE username=%s", (username,))
    cur_uid = c.fetchone()
    if cur_uid is None:
        raise Exception("The user " + username + " does not exist")
    return int(cur_uid[0])


def write_to_db(host_name, db_name, xml, new_pid, new_prid, username, status=1):
    con = get_connection(db_name, host_name)
    c = con.cursor()
    c.execute("SET search_path TO oabend")

    c.execute("SELECT id FROM users WHERE username=%s", (username,))
    cur_uid = c.fetchone()
    if cur_uid is None:
        raise Exception("The user " + username + " does not exist")
    else:
        cur_uid = cur_uid[0]
    now = datetime.datetime.now()
    c.execute("INSERT INTO xmls (reviewOf, xml, paid, prid, uid, comment, status, ts) "
              "VALUES (-1, %s, %s, %s, %s, %s, %s, %s) RETURNING id",
              (xml, new_pid, new_prid, cur_uid, '', status, now))
    queryset = c.fetchone()
    con.commit()
    return None if queryset is None else queryset[0]


def get_most_recent_xids(host_name, db_name, username):
    """Returns the most recent xids of the given username."""
    cur_uid = get_uid(host_name, db_name, username)
    c = get_cursor(host_name, db_name)
    c.execute("SELECT id, paid FROM xmls WHERE uid=%s ORDER BY ts DESC", (cur_uid,))
    print(username)
    print("=============")
    r = c.fetchone()
    count = 0
    while r and count < 10:
        print(r)
        r = c.fetchone()
        count += 1


def get_passage(host_name, db_name, pid):
    """Returns the passages with the given id numbers"""
    c = get_cursor(host_name, db_name)
    c.execute("SELECT passage FROM passages WHERE id=%s", (pid,))
    queryset = c.fetchone()
    if queryset is None:
        raise Exception("No passage with ID=" + pid)
    return queryset[0]


def linkage_type(u):
    """
    Returns the type of the primary linkage the scene participates in.
    It can be A,E or H. if it is a C, it returns the taf of the first fparent which is an A,E or H.
    If it does not find an fparent with either of these categories, it returns UNK_LINKAGE_TYPE.
    """
    cur_u = u
    while cur_u is not None:
        if cur_u.ftag in ['A', 'E', 'H']:
            return cur_u.ftag
        elif cur_u.ftag != 'C':
            return UNK_LINKAGE_TYPE
        else:
            cur_u = cur_u.fparent
    return UNK_LINKAGE_TYPE


def unit_length(u):
    """
    Returns the number of terminals (excluding remote units and punctuations) that are descendants of the unit u.
    """
    return len(u.get_terminals(punct=False, remotes=False))


def print_passages_to_file(host_name, db_name, paids, write_xml=False, write_site_xml=False, prefix='', start_index=0):
    """
    Returns for that user a list of submitted passages and a list of assigned but not submitted passages.
    Each passage is given in the format: (<passage ID>, <source>, <recent submitted xid or -1 if not submitted>,
    <number of tokens in the passage>, <number of units in the passage>, <number of scenes in the passage>,
    <average length of a scene>). It also returns a distribution of the categories.
    write_xml: determines whether to write it to a file, named <prefix><the number of the xml>.xml
    skip_first: the index of the passage where it should start looking (the ones before are skipped)
    """
    c = get_cursor(host_name, db_name)

    for paid in paids:
        if paid < start_index:  # skipping training passages
            continue
        c.execute("SELECT passage,source FROM passages WHERE id=%s", (paid,))
        r = c.fetchone()
        if r is not None:
            source = r[1]
            c.execute("SELECT id, xml,uid,ts FROM xmls WHERE paid=%s ORDER BY ts DESC", (paid,))
            r = c.fetchone()
            if r is not None:
                xid = r[0]
                uid = r[2]
                ts = r[3]
                print('\t'.join([str(paid), str(uid), str(source), str(xid), str(ts)]))

                if write_site_xml:
                    f = open(prefix + str(paid) + '_site.xml', 'w', encoding='utf-8')
                    f.write(r[1] + '\n')
                    f.close()
                # noinspection PyBroadException
                try:
                    ucca_dag = convert.from_site(fromstring(r[1]))
                except Exception:
                    sys.stderr.write("Skipped xid,paid " + str((xid, paid)) + "\n")
                    continue
                if write_xml:
                    f = open(prefix + str(paid) + '.xml', 'w')
                    f.write(tostring(convert.to_standard(ucca_dag)).decode())
                    f.close()


def get_predicates(host_name, db_name, only_complex=True):
    """
    Returns a list of all the predicates in the UCCA corpus.
    usernames -- the names of the users whose completed passages we should take.
    only_complex -- only the multi-word predicates will be returned.
    start_index -- the minimal passage number to be taken into account.
    """
    def _complex(u):
        """ Returns True if u is complex, i.e., if it has more than one child which is not an F or punct """
        if u is None or u.tag != 'FN':
            return False
        non_function_count = 0
        non_function_u = None
        for e in u.outgoing:
            if e.child.tag == 'FN' and e.tag != 'F':
                non_function_count += 1
                non_function_u = e.child
        return True if non_function_count > 1 else _complex(non_function_u)

    c = get_cursor(host_name, db_name)
    # uid = get_uid(host_name, db_name, username)
    # get all the completed xmls
    c.execute("SELECT id, xml FROM xmls WHERE status=%s AND reviewOf<>%s ORDER BY ts DESC", (1, -1))
    predicates = c.fetchall()

    with open('preds', 'w') as f:
        for r in tqdm(predicates):
            # noinspection PyBroadException
            try:
                ucca_dag = convert.from_site(fromstring(r[1]))
            except Exception:
                print("Skipped.", file=sys.stderr)
                continue

            # gathering statistics
            scenes = [x for x in ucca_dag.layer("1").all if x.tag == "FN" and x.is_scene()]
            for sc in scenes:
                main_relation = sc.process if sc.process is not None else sc.state
                if only_complex and not _complex(main_relation):
                    continue
                try:
                    print(main_relation.to_text(), file=f)
                except UnicodeEncodeError:
                    print("Skipped (encoding issue).", file=sys.stderr)
                    continue


def get_cursor(host_name, db_name):
    """ create a cursor to the search path """
    con = get_connection(db_name, host_name)
    c = con.cursor()
    c.execute("SET search_path TO oabend")
    return c


def get_connection(db_name, host_name):
    """ connects to the db and host, returns a connection object """
    global CONNECTION
    CONNECTION = psycopg2.connect(host=host_name, database=db_name)
    return CONNECTION


def main(argv):
    t = tqdm(globals()[argv[1]]("pgserver", "work", *argv[2:]), unit=" passages", desc="Downloading XMLs")
    for xml in t:
        p = convert.from_site(xml)
        t.set_postfix(ID=p.ID)
        convert.passage2file(p, p.ID + ".xml")


if __name__ == "__main__":
    main(sys.argv)



"""
The evaluation library for UCCA layer 1.
v1.3
2016-12-25: move common Fs to root before evaluation
2017-01-04: flatten centers, do not add 1 (for root) to mutual
2017-01-16: fix bug in moving common Fs
2018-04-12: exclude punctuation nodes regardless of edge tag
2018-12-11: fix another bug in moving common Fs
"""
from collections import Counter, OrderedDict
from itertools import groupby

from operator import attrgetter

from ucca import layer0, layer1, normalization
from ucca.constructions import get_by_names, create_passage_yields, PRIMARY, DEFAULT, ALL_EDGES
from ucca.layer1 import EdgeTags, NodeTags

UNLABELED = "unlabeled"
WEAK_LABELED = "weak_labeled"
LABELED = "labeled"

EVAL_TYPES = (LABELED, UNLABELED, WEAK_LABELED)

# Pairs that are considered as equivalent for the purposes of evaluation
EQUIV = ((EdgeTags.Process, EdgeTags.State),
         (EdgeTags.ParallelScene, EdgeTags.Center),
         (EdgeTags.Connector, EdgeTags.Linker),
         (EdgeTags.Function, EdgeTags.Relator))


def get_yield(unit):
    try:
        return frozenset(t.position for t in unit.get_terminals(punct=False))
    except ValueError:
        return frozenset()


def move_functions(p1, p2):
    """
    Move any common Fs to the root
    """
    f1, f2 = [{get_yield(u): u for u in p.layer(layer1.LAYER_ID).all
               if u.tag == NodeTags.Foundational and u.ftag == EdgeTags.Function} for p in (p1, p2)]
    for positions in f1.keys() & f2.keys():  # positions is a yield corresponding to a Function in both passages
        for (p, unit) in ((p1, f1[positions]), (p2, f2[positions])):
            unit.fparent.remove(unit)  # Remove from current primary parent (but preserve remote parents)
            p.layer(layer1.LAYER_ID).heads[0].add(EdgeTags.Function, unit)  # Add to root


def get_text(p, positions):
    l0 = p.layer(layer0.LAYER_ID)
    return [l0.by_position(i).text for i in range(1, len(l0.all) + 1) if i in positions]


def print_tags_and_text(p, yield_tags):
    text_to_tags = {}
    for construction, construction_yield_tags in yield_tags.items():
        for y, tags in construction_yield_tags.items():
            if construction.criterion is None:  # category from reference yield tags
                tags = list(tags) + [construction.name]
            text_to_tags.setdefault((min(y or [-1]), -max(y or [-1]), " ".join(get_text(p, y))), []).extend(tags)
    for (_, _, text), tags in sorted(text_to_tags.items()):
        print((",".join(sorted(set(filter(None, tags)))) + ": " + text) if tags else text)


def expand_equivalents(tag_set):
    """
    Returns a set of all the tags in the tag set or those equivalent to them
    :param tag_set: set of tags (strings) to expand
    """
    return tag_set.union(t1 for t in tag_set for pair in EQUIV for t1 in pair if t in pair and t != t1)


class Evaluator:
    def __init__(self, verbose, constructions, units, fscore, errors):
        """
        :param verbose: whether to print the scores
        :param constructions: names of construction types to include in the evaluation
        :param units: whether to calculate and print the mutual and exclusive units in the passages
        :param fscore: whether to find and return the scores
        :param errors: whether to calculate and print the confusion matrix of errors
        """
        self.verbose = verbose
        self.constructions = list(DEFAULT.values()) + [c for c in get_by_names(constructions)
                                                       if c not in DEFAULT.values()]
        self.units = units
        self.fscore = fscore
        self.errors = errors

    @staticmethod
    def find_mutuals(m1, m2, eval_type, mutual_tags, counter=None):
        for y in m1.keys() & m2.keys():
            if eval_type == UNLABELED:
                mutual_tags[y] = ()
            else:
                tags = [set(c.edge.tag for c in m[y]) for m in (m1, m2)]
                if eval_type == WEAK_LABELED:
                    tags[0] = expand_equivalents(tags[0])
                intersection = set.intersection(*tags)
                if intersection:  # non-empty intersection
                    mutual_tags[y] = intersection
        if counter is not None:
            for y in m1.keys() | m2.keys():
                tags = [sorted(set(c.edge.tag for c in m.get(y, ()) if not c.is_unary_child)) for m in (m1, m2)]
                counter[tuple("|".join(t) or "<UNMATCHED>" for t in tags)] += 1

    def get_scores(self, p1, p2, eval_type, r=None):
        """
        prints the relevant statistics and f-scores. eval_type can be 'unlabeled', 'labeled' or 'weak_labeled'.
        calculates a set of all the yields such that both passages have a unit with that yield.
        :param p1: passage to compare
        :param p2: reference passage object
        :param eval_type: evaluation type to use, out of EVAL_TYPES
        1. UNLABELED: it doesn't matter what labels are there.
        2. LABELED: also requires tag match (if there are multiple units with the same yield, requires one match)
        3. WEAK_LABELED: also requires weak tag match (if there are multiple units with the same yield,
                         requires one match)
        :param r: reference passage for fine-grained evaluation
        :returns: EvaluatorResults object if self.fscore is True, otherwise None
        """
        mutual = OrderedDict()
        counters = OrderedDict() if self.errors and eval_type == LABELED else None
        passage_yields = create_passage_yields(r or p2)
        reference_yield_tags = passage_yields[ALL_EDGES.name] if passage_yields else None
        maps = [{} if p is None else create_passage_yields(p, self.constructions, tags=False, reference=p2,
                                                           reference_yield_tags=reference_yield_tags) for p in (p1, p2)]
        if p1 is not None:
            ordered_constructions = [c for c in self.constructions if any(c in m for m in maps)]
            for m in maps[::-1]:
                ordered_constructions += [c for c in m if c not in ordered_constructions]
            for construction in ordered_constructions:
                yield_cands = [m.get(construction, {}) for m in maps]
                self.find_mutuals(*yield_cands, eval_type=eval_type, mutual_tags=mutual.setdefault(construction, {}),
                                  counter=None if counters is None else counters.setdefault(construction, Counter()))

        only = [{c: {y: tags for y, tags in d.items() if y not in mutual[c]} for c, d in m.items()} for m in maps]
        res = EvaluatorResults((c, SummaryStatistics(len(mutual[c]), len(only[0].get(c, ())), len(only[1].get(c, ())),
                                                     None if counters is None else counters.get(c))) for c in mutual)
        if self.verbose:
            print("Evaluation type: (" + eval_type + ")")
            if self.units and p1 is not None:
                print("==> Mutual Units:")
                print_tags_and_text(p1, mutual)
                print("==> Only in guessed:")
                print_tags_and_text(p1, only[0])
                print("==> Only in reference:")
                print_tags_and_text(p2, only[1])
            if self.fscore:
                res.print()
        return res


class Scores:
    def __init__(self, evaluator_results, name=None, evaluation_format=None):
        """
        :param evaluator_results: dict: eval_type -> EvaluatorResults
        :param name: if not UCCA, name of evaluated format
        :param evaluation_format: if not ucca, lowercase string representation of evaluated format
        """
        self.evaluators = dict(evaluator_results)
        self.name = name or "UCCA"
        self.format = evaluation_format or "ucca"

    def average_f1(self, mode=LABELED):
        """
        Calculate the average F1 score across primary and remote edges
        :param mode: LABELED, UNLABELED or WEAK_LABELED
        :return: a single number, the average F1
        """
        return float(self[mode].aggregate_default().f1)

    @staticmethod
    def aggregate(scores):
        """
        Aggregate multiple Scores instances
        :param scores: iterable of Scores
        :return: new Scores with aggregated scores
        """
        scores = list(scores)
        evaluators = [s.evaluators for s in scores]
        names = list(set(s.name for s in scores))
        formats = list(set(s.format for s in scores))
        return Scores(((t, EvaluatorResults.aggregate(filter(None, (e.get(t) for e in evaluators))))
                       for t in EVAL_TYPES),
                      name=names[0] if len(names) == 1 else None,
                      evaluation_format=formats[0] if len(formats) == 1 else None)

    def print(self, eval_type=None, **kwargs):
        for eval_type in EVAL_TYPES if eval_type is None else [eval_type]:
            evaluator = self.evaluators.get(eval_type)
            if evaluator:
                print("Evaluation type: (" + eval_type + ")", **kwargs)
                evaluator.print(**kwargs)

    def print_confusion_matrix(self, *args, eval_type=None, **kwargs):
        for eval_type in EVAL_TYPES if eval_type is None else [eval_type]:
            evaluator = self.evaluators.get(eval_type)
            if evaluator:
                evaluator.print_confusion_matrix("Evaluation type: (" + eval_type + ")", *args, **kwargs)

    def fields(self, eval_type=LABELED, counts=False):
        e = self[eval_type]
        attrs = ("num_guessed", "num_ref", "num_matches") if counts else ("p", "r", "f1")
        return ["%.3f" % float(getattr(x, y)) for x in e.results.values() for y in attrs]

    def titles(self, eval_type=LABELED, counts=False):
        return self.field_titles(self[eval_type].results.keys(), eval_type=eval_type, counts=counts)

    @staticmethod
    def field_titles(constructions=DEFAULT, eval_type=LABELED, counts=False):
        titles = ("guessed", "ref", "matches") if counts else ("precision", "recall", "f1")
        return ["_".join(((str(x),) if len(constructions) > 1 else ()) + (eval_type, y))
                for x in constructions for y in titles]

    def __getitem__(self, eval_type):
        return self.evaluators[eval_type]


class EvaluatorResults:
    def __init__(self, results, default=None):
        """
        :param results: dict: Construction -> SummaryStatistics
        :param default: map of default constructions (default is primary and remote)
        """
        self.default = default or DEFAULT
        self.results = OrderedDict(results)
        self.results.update(((c, self[c]) for c in self.default.values()))  # Make sure there are entries for defaults

    def print(self, **kwargs):
        for construction, stats in self.results.items():
            if len(self.results) > 1:
                print("\n%s:" % construction.description, **kwargs)
            stats.print(**kwargs)
        print(**kwargs)

    def print_confusion_matrix(self, prefix=None, sep=None, as_table=False, **kwargs):
        primary = self[PRIMARY]
        if primary.errors:
            errors = primary.errors.most_common()
            if sep:
                print(sep.join(("guessed", "ref", "count")), **kwargs)
                for error, freq in errors:
                    print(sep.join(error + (str(freq),)), **kwargs)
            else:
                print("\n%sConfusion Matrix:" % ("" if prefix is None else (prefix + ", ")), **kwargs)
                if as_table:
                    y_labels = sorted(set(x[0][1] for x in errors))
                    print("", *y_labels, sep="\t")
                    for x, x_errors in groupby(sorted(errors), key=lambda x: x[0][0]):
                        errors_by_y = Counter()
                        for (_, y), f in x_errors:
                            errors_by_y[y] += f
                        print(x, *[errors_by_y.get(y, 0) for y in y_labels], sep="\t")
                else:
                    for error, freq in errors:
                        l1 = max(len(e1) for e1, _ in primary.errors)
                        l2 = max(len(e2) for _, e2 in primary.errors)
                        print("%-*s %-*s %d" % (l1, error[0], l2, error[1], freq), **kwargs)

    @classmethod
    def aggregate(cls, results):
        """
        :param results: iterable of EvaluatorResults
        :return: new EvaluatorResults with aggregates scores
        """
        collected = OrderedDict()
        default = OrderedDict()
        for evaluator_results in results:
            for c, r in evaluator_results.results.items():
                collected.setdefault(c, []).append(r)
            default.update(evaluator_results.default)
        return EvaluatorResults(((c, SummaryStatistics.aggregate(r)) for c, r in collected.items()), default=default)

    def aggregate_default(self):
        """
        Aggregate primary and remote SummaryStatistics in this EvaluatorResults instance
        :return: SummaryStatistics object representing aggregation over primary and remote
        """
        return SummaryStatistics.aggregate([self[c] for c in self.default.values()])

    def __bool__(self):
        return bool(self.results and any(self.results.values()))

    def __getitem__(self, construction):
        return self.results.get(construction, SummaryStatistics(0, 0, 0, Counter()))


class SummaryStatistics:
    def __init__(self, num_matches, num_only_guessed, num_only_ref, errors=None):
        self.num_matches = num_matches
        self.num_only_guessed = num_only_guessed
        self.num_only_ref = num_only_ref
        self.num_guessed = num_matches + num_only_guessed
        self.num_ref = num_matches + num_only_ref
        self.p = 1.0 if self.num_guessed == 0 else 1.0 * num_matches / self.num_guessed
        self.r = 1.0 if self.num_ref == 0 else 1.0 * num_matches / self.num_ref
        self.f1 = 0.0 if 0.0 in (self.p, self.r) else 2.0 * self.p * self.r / float(self.p + self.r)
        self.errors = errors

    def print(self, **kwargs):
        print("Precision: {:.3} ({}/{})".format(self.p, self.num_matches, self.num_guessed), **kwargs)
        print("Recall: {:.3} ({}/{})".format(self.r, self.num_matches, self.num_ref), **kwargs)
        print("F1: {:.3}".format(self.f1), **kwargs)

    @classmethod
    def aggregate(cls, stats):
        """
        :param stats: iterable of SummaryStatistics
        :return: new SummaryStatistics with aggregated scores
        """
        return SummaryStatistics(*map(sum, [map(attrgetter(attr), stats)
                                            for attr in ("num_matches", "num_only_guessed", "num_only_ref")]),
                                 Counter({k: sum((s.errors or {}).get(k, 0) for s in stats)
                                          for k in set.union(*[set(s.errors or ()) for s in stats])}))

    def __bool__(self):
        return bool(self.num_matches or self.num_only_guessed or self.num_only_ref or self.errors)


def evaluate(guessed, ref, converter=None, verbose=False, constructions=DEFAULT,
             units=False, fscore=True, errors=False, normalize=True, eval_type=None, ref_yield_tags=None, **kwargs):
    """
    Compare two passages and return requested diagnostics and scores, possibly printing them too.
    NOTE: since normalize=True by default, this method is destructive: it modifies the given passages before evaluation.
    :param guessed: Passage object to evaluate
    :param ref: reference Passage object to compare to
    :param converter: optional function to apply to passages before evaluation
    :param verbose: whether to print the results
    :param constructions: names of construction types to include in the evaluation
    :param units: whether to evaluate common units
    :param fscore: whether to compute precision, recall and f1 score
    :param errors: whether to print the mistakes
    :param normalize: flatten centers and move common functions to root before evaluation - modifies passages
    :param eval_type: specific evaluation type to limit to
    :param ref_yield_tags: reference passage for fine-grained evaluation
    :return: Scores object
    """
    del kwargs
    if converter is not None:
        guessed = converter(guessed)
        ref = converter(ref)
    if normalize:  # FIXME clone passages to avoid modifying the original ones
        for passage in (guessed, ref):
            normalization.normalize(passage)  # flatten Cs inside Cs
        move_functions(guessed, ref)  # move common Fs to be under the root

    evaluator = Evaluator(verbose, constructions, units, fscore, errors)
    return Scores((evaluation_type, evaluator.get_scores(guessed, ref, evaluation_type, r=ref_yield_tags))
                  for evaluation_type in ([eval_type] if eval_type else EVAL_TYPES))



#!/usr/bin/env python3

import argparse
import os
from xml.etree.ElementTree import tostring

from ucca import convert
from ucca.ioutil import external_write_mode
from ucca.ioutil import get_passages_with_progress_bar

desc = """Parses an XML in UCCA standard format, and writes them in old site format."""


def main(args):
    os.makedirs(args.outdir, exist_ok=True)
    for passage in get_passages_with_progress_bar(args.filenames):
        site_filename = os.path.join(args.outdir, passage.ID + ".xml")
        with open(site_filename, "w", encoding="utf-8") as f:
            print(tostring(convert.to_site(passage)).decode(), file=f)
        if args.verbose:
            with external_write_mode():
                print("Wrote '%s'" % site_filename)


if __name__ == "__main__":
    argparser = argparse.ArgumentParser(description=desc)
    argparser.add_argument("filenames", nargs="+", help="XML file names to convert")
    argparser.add_argument("-o", "--outdir", default=".", help="output directory")
    argparser.add_argument("-v", "--verbose", action="store_true", help="verbose output")
    main(argparser.parse_args())
